"""
PyDetex
https://github.com/ppizarror/PyDetex

GUI
Basic gui that convers and executes a given pipeline.
"""

__all__ = ['PyDetexGUI']

import concurrent.futures
import tkinter as tk
from tkinter import filedialog
from tkinter import ttk
from tkinter import messagebox

import os
import pyperclip
import requests
import string
import sys
import traceback

from nltk.tokenize import RegexpTokenizer
from outdated import check_outdated
from PyMultiDictionary import MultiDictionary
from typing import Optional, Tuple
from warnings import warn

sys.path.append('../')

import pydetex._gui_utils as gui_ut
import pydetex.parsers as par
import pydetex.pipelines as pip
import pydetex.utils as ut
import pydetex.version

from pydetex._fonts import FONT_TAGS
from pydetex._gui_settings import Settings

# Settings
_MAX_PASTE_RETRY: int = 3


class PyDetexGUI(object):
    """
    GUI.
    """

    _cfg: 'Settings'
    _clear_button: 'ut.Button'
    _clip: bool
    _copy_clip_button: 'tk.Button'
    _detect_language_event_id: str
    _detected_lang_tag: str
    _dictionary: 'MultiDictionary'
    _dictionary_btn: 'tk.Button'
    _dictionary_window: Optional['gui_ut.DictionaryGUI']
    _paste_timeout_error: int
    _process_button: 'ut.Button'
    _process_clip_button: 'tk.Button'
    _processing: bool
    _ready: bool
    _root: 'tk.Tk'
    _settings_window: Optional['gui_ut.SettingsWindow']
    _status_bar_cursor: 'tk.Label'
    _status_bar_cursor_sel: 'tk.Label'
    _status_bar_lang: 'tk.Label'
    _status_bar_status: 'tk.Label'
    _status_bar_words: 'tk.Label'
    _status_clear_event_id: str
    _text_in: 'gui_ut.RichText'
    _text_out: 'gui_ut.RichText'
    _tokenizer: 'RegexpTokenizer'

    def __init__(self) -> None:
        """
        Constructor.
        """
        # ----------------------------------------------------------------------
        # Creates the window
        # ----------------------------------------------------------------------
        self._root = tk.Tk()

        # Dictionary
        self._dictionary = MultiDictionary()
        self._dictionary_window = None

        # Load settings
        self._cfg = Settings()
        window_size = self._cfg.get(self._cfg.CFG_WINDOW_SIZE).copy()

        # Configure window
        self._root.title('PyDetex')
        img = tk.Image('photo', file=ut.RESOURCES_PATH + 'icon.gif')
        # noinspection PyProtectedMember,PyUnresolvedReferences
        self._root.tk.call('wm', 'iconphoto', self._root._w, img)
        if not ut.IS_OSX:
            try:
                self._root.iconbitmap(ut.RESOURCES_PATH + 'icon.ico')
            except tk.TclError:  # Linux
                pass

        self._root.minsize(width=window_size[0], height=window_size[1])
        self._root.resizable(width=False, height=False)
        gui_ut.center_window(self._root, window_size)
        self._root.protocol('WM_DELETE_WINDOW', self._close)

        # ----------------------------------------------------------------------
        # Settings button
        # ----------------------------------------------------------------------
        f0 = tk.Frame(self._root, border=10, width=window_size[0], height=50)
        f0.pack()
        f0.pack_propagate(False)
        tk.Button(f0, text=ut.button_text(self._cfg.lang('open_file')), command=self._open_file,
                  relief=tk.GROOVE).pack(side=tk.LEFT)
        tk.Button(f0, text=ut.button_text(self._cfg.lang('about')), command=self._about,
                  relief=tk.GROOVE).pack(side=tk.RIGHT)
        tk.Button(f0, text=ut.button_text(self._cfg.lang('settings')), command=self._open_settings,
                  relief=tk.GROOVE).pack(side=tk.RIGHT, padx=(0, 7 if ut.IS_OSX else 11))
        self._dictionary_btn = tk.Button(f0, text=ut.button_text(self._cfg.lang('dictionary')),
                                         command=self._open_dictionary, relief=tk.GROOVE)
        self._dictionary_btn.pack(side=tk.RIGHT, padx=(0, 7 if ut.IS_OSX else 11))

        # ----------------------------------------------------------------------
        # Input texts
        # ----------------------------------------------------------------------
        hthick, hcolor = 3 if ut.IS_OSX else 1, '#426392' if ut.IS_OSX else '#475aff'
        fsize = self._cfg.get(self._cfg.CFG_FONT_SIZE)
        show_lnum = self._cfg.get(self._cfg.CFG_SHOW_LINE_NUMBERS)
        self._tab_spaces = 4

        # In text
        f1 = tk.Frame(self._root, border=0, width=window_size[0], height=window_size[2])
        f1.pack(fill='both', padx=10)
        f1.pack_propagate(False)

        self._text_in = gui_ut.RichText(self._cfg, self._root, f1, wrap='word', highlightthickness=hthick,
                                        highlightcolor=hcolor, font_size=fsize, editable=True,
                                        scrollbar_y=f1, add_line_numbers=f1 if show_lnum else None)
        self._text_in.pack(fill='both')
        self._text_in.bind('<Button>', self._process_cursor_in)
        self._text_in.bind('<ButtonRelease>', self._process_cursor_in)
        self._text_in.bind('<FocusIn>', self._process_focusin_in)
        self._text_in.bind('<FocusOut>', self._process_focusout_in)
        # noinspection PyTypeChecker
        self._text_in.bind('<Key>', self._process_in_key, add='+')
        self._text_in.bind('<Key-Tab>', self._text_in.tab_selected)
        self._text_in.bind('<Shift-KeyPress-Tab>', self._text_in.undo_tab_selected)
        self._text_in.tab_spaces = self._tab_spaces

        # Out text
        f2 = tk.Frame(self._root, border=0, width=window_size[0], height=window_size[2])
        f2.pack(fill='both', padx=10, pady=(window_size[3], 0))
        f2.pack_propagate(False)

        self._text_out = gui_ut.RichText(self._cfg, self._root, f2, wrap='word', highlightthickness=hthick,
                                         highlightcolor=hcolor, font_size=fsize, copy=True,
                                         scrollbar_y=f2, add_line_numbers=f2 if show_lnum else None)
        # noinspection PyTypeChecker
        self._text_out.bind('<Key>', self._process_out_key, add='+')
        self._text_out.pack(fill='both')

        # ----------------------------------------------------------------------
        # Commands buttons
        # ----------------------------------------------------------------------
        command_btn_packx = 15  # margin px
        f3 = tk.Frame(self._root, border=2)
        f3.pack(pady=(window_size[4], window_size[4] + 5))

        # Process
        self._process_button = ut.Button(f3, text=ut.button_text(self._cfg.lang('process')),
                                         command=self._process, relief=tk.GROOVE,
                                         bg='#475aff' if ut.IS_OSX else '#6388ff')
        self._process_button.pack(side=tk.LEFT, padx=(0, command_btn_packx))

        # Process clip
        self._process_clip_button = tk.Button(f3, text=ut.button_text(self._cfg.lang('process_clip')),
                                              command=self._process_clip, relief=tk.GROOVE)
        self._process_clip_button.pack(side=tk.LEFT, padx=(0, command_btn_packx))

        # Copy to clip
        self._copy_clip_button = tk.Button(f3, text=ut.button_text(self._cfg.lang('process_copy')),
                                           command=self._copy_to_clip, relief=tk.GROOVE)
        self._copy_clip_button.pack(side=tk.LEFT, padx=(0, command_btn_packx))

        # Clear
        self._clear_button = ut.Button(f3, text=ut.button_text(self._cfg.lang('clear')), command=self._clear,
                                       relief=tk.GROOVE, bg='#ff7878')
        self._clear_button.pack(side=tk.LEFT)

        # ----------------------------------------------------------------------
        # Status bar
        # ----------------------------------------------------------------------
        ttk.Separator(self._root, orient='horizontal').pack(side='top', fill='x')
        status_fg = '#999999' if ut.IS_OSX else '#666666'

        f4 = tk.Frame(self._root, width=window_size[0], height=26)
        f4.pack(fill='both')
        f4.pack_propagate(False)

        # Detected language
        show_status = 0.2 if window_size[0] > 750 else 0
        self._status_bar_lang = gui_ut.make_label(f4, w=window_size[0] * (0.4 + (0.2 - show_status)), h=20,
                                                  side=tk.LEFT, fg=status_fg, bd=0, relief=tk.SUNKEN, anchor=tk.W,
                                                  pad=(0, 0, 0, 10), separator=True)

        # Status
        self._status_bar_status = gui_ut.make_label(f4, w=window_size[0] * show_status, h=20, side=tk.LEFT,
                                                    fg=status_fg, bd=0, relief=tk.SUNKEN, anchor=tk.W, pad=(0, 0, 0, 5),
                                                    separator=True)

        # Cursor
        self._status_bar_cursor = gui_ut.make_label(f4, w=window_size[0] * 0.11, h=20, side=tk.LEFT, fg=status_fg,
                                                    bd=0, relief=tk.SUNKEN, anchor=tk.W, pad=(0, 0, 0, 5))

        # Cursor selected
        self._status_bar_cursor_sel = gui_ut.make_label(f4, w=window_size[0] * 0.14, h=20, side=tk.LEFT, fg=status_fg,
                                                        bd=0, relief=tk.SUNKEN, anchor=tk.W, pad=(0, 0, 0, 5),
                                                        separator=True)

        # Total processed words
        self._status_bar_words = gui_ut.make_label(f4, w=window_size[0] * 0.15, h=20, side=tk.LEFT, fg=status_fg,
                                                   bd=0, relief=tk.SUNKEN, anchor=tk.W, pad=(0, 0, 0, 5))

        # ----------------------------------------------------------------------
        # Final settings
        # ----------------------------------------------------------------------
        self._status_clear_event_id = ''
        self._clear()  # This also changes states

        # Check if pyperclip is enabled
        self._clip = True
        try:
            pyperclip.paste()
        except pyperclip.PyperclipException:
            self._clip = False
            error = 'pyperclip is not available on your system (copy/paste mechanism). GUI buttons were disabled'
            warn(error)
            self._process_clip_button['state'] = tk.DISABLED
            self._copy_clip_button['state'] = tk.DISABLED

        # Set variables
        self._detect_language_event_id = ''
        self._detected_lang_tag = '–'
        self._paste_timeout_error = 0
        self._processing = False
        self._ready = False
        self._settings_window = None
        self._tokenizer = RegexpTokenizer(r'\w+')

        # Inserts the placeholder text
        self._insert_in(self._cfg.lang('placeholder'))

        # Finals
        self._text_out['state'] = tk.DISABLED
        self._root.update()

    def _open_file(self) -> None:
        """
        Opens a file.
        """
        self._status(self._cfg.lang('status_requesting_file'))
        last_path_cfg = self._cfg.get(self._cfg.CFG_LAST_OPENED_FOLDER)
        initial_dir = '/' if not os.path.isdir(last_path_cfg) else last_path_cfg
        filename = filedialog.askopenfilename(
            title=self._cfg.lang('open_file_select'),
            initialdir=initial_dir,
            filetypes=[(self._cfg.lang('open_file_latex_file'), '*.tex')])
        if filename == '':
            return self._status_clear()
        try:
            filename_dir = os.path.dirname(filename)
        except TypeError:  # Linux
            return
        self._cfg.set(self._cfg.CFG_LAST_OPENED_FOLDER, filename_dir)
        try:
            text = ut.open_file(filename)
            self._clear()
            self._insert_in(text)
            self._cfg.save()
            os.chdir(filename_dir)
        except PermissionError:
            pass

    def _insert_in(self, text: str, clear: bool = False) -> None:
        """
        Insert text to in widget.

        :param text: Text
        :param clear: Clear the text
        """
        if clear:
            self._text_in.clear()
        text = text.replace('\t', ' ' * self._tab_spaces)
        self._text_in.insert(tk.END, text)
        self._text_in.redraw()
        self._detect_language()
        self._process_cursor_in(None)

    def _status(self, text: str, clear: bool = False, clear_time: int = 500) -> None:
        """
        Set the app status.

        :param text: Status text
        :param clear: Clear after time
        :param clear_time: Clear time in miliseconds
        """
        assert clear_time > 0
        if self._status_clear_event_id != '':
            self._root.after_cancel(self._status_clear_event_id)
        self._status_bar_status['text'] = text
        if clear:
            self._status_clear_event_id = self._root.after(clear_time, self._status_clear)

    def _status_clear(self) -> None:
        """
        Clear the status text.
        """
        self._status(self._cfg.lang('status_idle'))

    # noinspection PyUnresolvedReferences
    def _process_in_key(self, event: 'tk.Event') -> Optional['tk.Event']:
        """
        Process in keys.

        :param event: Event
        :return: Event
        """
        if self._detect_language_event_id != '':
            self._root.after_cancel(self._detect_language_event_id)

        if event.char in string.printable and event.char != '':
            self._status(self._cfg.lang('status_writing'), True, 1000)
        self._detect_language_event_id = self._root.after(100, self._detect_language)
        self._process_cursor_in(event)
        return event

    def _process_cursor_in(self, event: Optional['tk.Event']) -> Optional['tk.Event']:
        """
        Process cursor on input text.

        :param event: Event
        :return: Event
        """
        self._root.after(50, self._process_cursor_event)
        return event

    def _process_focusin_in(self, event: Optional['tk.Event']) -> Optional['tk.Event']:
        """
        Process focus in on input text.

        :param event: Event
        :return: Event
        """
        self._process_cursor_in(event)
        return event

    def _process_focusout_in(self, event: Optional['tk.Event']) -> Optional['tk.Event']:
        """
        Process focus out on input text.

        :param event: Event
        :return: Event
        """
        fout = self._cfg.lang('status_cursor_input_focusout')
        w = self._cfg.get(self._cfg.CFG_WINDOW_SIZE)[0]
        if w < 1000:
            fout = self._cfg.lang('status_cursor_input_focusout_min')
        if w < 750:
            fout = self._cfg.lang('status_cursor_input_focusout_min2')
        self._status_bar_cursor['text'] = fout
        self._status_bar_cursor_sel['text'] = ''
        return event

    def _process_cursor_event(self) -> None:
        """
        Process cursor position.
        """
        window_w = self._cfg.get(self._cfg.CFG_WINDOW_SIZE)[0]
        cur_pos = self._text_in.index(tk.INSERT)

        # Check cursor position
        if cur_pos is None:
            self._status_bar_cursor['text'] = self._cfg.lang('status_cursor_null')
        else:
            line, pos = tuple(cur_pos.split('.'))
            cur_t = self._cfg.lang('status_cursor') if window_w > 750 else self._cfg.lang('status_cursor_min')
            self._status_bar_cursor['text'] = cur_t.format(pos, line)

        # Get cursor selection
        try:
            sf = self._text_in.count('1.0', 'sel.first')
            sl = self._text_in.count('1.0', 'sel.last')
            if sf is None:
                sf = (0,)
            d = sl[0] - sf[0]
            sel = self._cfg.lang('status_cursor_selected') if window_w > 750 else self._cfg.lang(
                'status_cursor_selected_min')
            chars = self._cfg.lang('status_cursor_selected_chars') if window_w > 900 else self._cfg.lang(
                'status_cursor_selected_chars_min')
            s = f'{sel}: '
            text = self._text_in.get(0.0, tk.END)
            if d == 1:
                self._status_bar_cursor_sel['text'] = \
                    s + self._cfg.lang('status_cursor_selected_chars_single')
            elif d == len(text) or d == len(text.strip()):
                self._status_bar_cursor_sel['text'] = self._cfg.lang('status_cursor_selected_all')
            else:
                self._status_bar_cursor_sel['text'] = s + chars.format(d)

        except tk.TclError:
            self._status_bar_cursor_sel['text'] = ''

    @staticmethod
    def _process_out_key(event: 'tk.Event') -> Optional['tk.Event']:
        """
        Process out keys.

        :param event: Event
        :return: Event
        """
        # noinspection PyUnresolvedReferences
        if event.char == '':
            return event

    def _detect_language(self) -> None:
        """
        Detects the input lang.
        """
        text = self._text_in.get(0.0, tk.END).strip()

        # Check language
        self._detected_lang_tag = ut.detect_language(text)
        lang = self._dictionary.get_language_name(self._detected_lang_tag, self._cfg.get(self._cfg.CFG_LANG))
        if text != '':
            self._status_bar_lang['text'] = self._cfg.lang('detected_lang').format(lang, self._detected_lang_tag)
        else:
            self._status_bar_lang['text'] = self._cfg.lang('detected_lang_write')
        self._detect_language_event_id = ''

        # Check if dictionary is available
        # noinspection PyProtectedMember
        if self._detected_lang_tag in self._dictionary._langs.keys():
            self._status_bar_lang['text'] += f' [{self._cfg.lang("dictionary")}]'
            self._dictionary_btn['state'] = tk.NORMAL
        else:
            self._dictionary_btn['state'] = tk.DISABLED

    def start(self) -> None:
        """
        Starts the application.
        """
        # noinspection PyProtectedMember
        if self._cfg._last_opened_day_diff >= 7:
            self._root.after(1000, self._check_version_event)
        self._root.after(100, lambda: self._root.lift())
        self._text_in.redraw()
        self._text_out.redraw()
        self._root.mainloop()

    def _clear(self) -> None:
        """
        Clear texts.
        """
        self._text_out['state'] = tk.NORMAL
        self._text_in.delete(0.0, tk.END)
        self._text_out.delete(0.0, tk.END)
        self._text_out['state'] = tk.DISABLED
        self._copy_clip_button['state'] = tk.DISABLED

        self._ready = False
        self._detect_language()
        self._status_bar_words['text'] = self._cfg.lang('status_words').format(0)

        self._status_clear()
        self._process_cursor_event()
        self._text_in.focus_force()
        self._text_in.redraw()
        self._text_out.redraw()

    @property
    def pipeline(self) -> pip.PipelineType:
        """
        Return the pipeline.

        :return: Pipeline
        """
        return self._cfg.get(self._cfg.CFG_PIPELINE)

    def _process(self) -> None:
        """
        Process and call the pipeline.
        """
        if self._processing:
            return
        self._status(self._cfg.lang('status_processing'), True)
        self._processing = True
        # for btn in (self._process_button, self._process_clip_button,
        #             self._copy_clip_button, self._clear_button):
        #     btn['state'] = tk.DISABLED
        try:
            self._root['cursor'] = 'wait'
        except tk.TclError:
            pass
        self._root.after(1, lambda: self._process_inner())

    def _process_inner(self) -> None:
        """
        Process called after.
        """
        text = self._text_in.get(0.0, tk.END)
        self._text_out['state'] = tk.NORMAL
        if self._clip:
            self._copy_clip_button['state'] = tk.NORMAL

        # Font format
        font_format = self._cfg.get(self._cfg.CFG_OUTPUT_FONT_FORMAT)
        par.FONT_FORMAT_SETTINGS['bold'] = FONT_TAGS['bold'] if font_format else ''
        par.FONT_FORMAT_SETTINGS['cite'] = FONT_TAGS['link'] if font_format else ''
        par.FONT_FORMAT_SETTINGS['equation'] = FONT_TAGS['equation_inside'] if font_format else ''
        par.FONT_FORMAT_SETTINGS['hl'] = FONT_TAGS['highlight'] if font_format else ''
        par.FONT_FORMAT_SETTINGS['italic'] = FONT_TAGS['italic'] if font_format else ''
        par.FONT_FORMAT_SETTINGS['normal'] = FONT_TAGS['normal'] if font_format else ''
        par.FONT_FORMAT_SETTINGS['ref'] = FONT_TAGS['link'] if font_format else ''
        par.FONT_FORMAT_SETTINGS['strike'] = FONT_TAGS['strike'] if font_format else ''
        par.FONT_FORMAT_SETTINGS['tex_text_tag'] = FONT_TAGS['bold'] if font_format else ''
        par.FONT_FORMAT_SETTINGS['tex_text_tag_content'] = FONT_TAGS['italic'] if font_format else ''
        par.FONT_FORMAT_SETTINGS['underline'] = FONT_TAGS['underlined'] if font_format else ''

        # Process the text and get the language
        # noinspection PyBroadException
        try:
            out = self.pipeline(
                text,
                self._detected_lang_tag,
                show_progress=True,
                replace_defs=self._cfg.get(self._cfg.CFG_PIPELINE_REPLACE_DEFS),
                replace_pydetex_tag_dollar_symbol=False,  # Avoid highlight problems
                compress_cite=self._cfg.get(self._cfg.CFG_PIPELINE_COMPRESS_CITE)
            )
        except Exception:
            err = self._cfg.lang('process_error').format(
                pydetex.__url_bug_tracker__,
                FONT_TAGS['error'] + traceback.format_exc()
            )
            self._text_out.insert_highlighted_text(FONT_TAGS['normal'] + err, True)
            return self._process_final()
        words = len(self._tokenizer.tokenize(out))
        self._cfg.add_words(words)

        # Check repeated words
        if self._cfg.get(self._cfg.CFG_CHECK_REPETITION):
            out = ut.check_repeated_words(
                s=out,
                lang=self._detected_lang_tag,
                min_chars=self._cfg.get(self._cfg.CFG_REPETITION_MIN_CHAR),
                window=self._cfg.get(self._cfg.CFG_REPETITION_DISTANCE),
                stopwords=self._cfg.get(self._cfg.CFG_REPETITION_USE_STOPWORDS),
                stemming=self._cfg.get(self._cfg.CFG_REPETITION_USE_STEMMING),
                ignore=self._tokenizer.tokenize(self._cfg.get(self._cfg.CFG_REPETITION_IGNORE_WORDS)),
                font_tag_format=FONT_TAGS['repeated_tag'] if font_format else '',
                font_param_format=FONT_TAGS['repeated_word'] if font_format else '',
                font_normal_format=FONT_TAGS['normal'] if font_format else '',
                remove_tokens=list(FONT_TAGS.values()),
                tag=self._cfg.lang('tag_repeated')
            )

        # Apply syntax highlight
        out = par.replace_pydetex_tags(ut.syntax_highlight(out))

        # Insert the text
        self._text_out.insert_highlighted_text(out, True, font_format)

        # Final
        self._process_final(words)

    def _process_final(self, words: int = 0) -> None:
        """
        Function executed after process finished.

        :param words: Total processed words
        """
        # Configure status
        try:
            self._root['cursor'] = ''
        except tk.TclError:
            pass
        self._process_button['state'] = tk.NORMAL
        self._clear_button['state'] = tk.NORMAL
        if self._clip:
            self._process_clip_button['state'] = tk.NORMAL
            self._copy_clip_button['state'] = tk.NORMAL
        self._processing = False
        self._ready = True
        self._status_bar_words['text'] = self._cfg.lang('status_words').format(words)
        self._text_out['state'] = tk.DISABLED
        self._text_out.redraw()

        # Detect language to rewrite the status
        self._detect_language()

        # If auto copy
        if self._cfg.get(self._cfg.CFG_PROCESS_AUTO_COPY):
            self._copy_to_clip()

    def _process_clip(self) -> None:
        """
        Process from clipboard. Tries with pooling.
        """
        if not self._clip:
            return

        def _paste():
            return pyperclip.paste()

        self._status(self._cfg.lang('copy_from_clip'), True)
        executor = concurrent.futures.ThreadPoolExecutor(1)
        future = executor.submit(_paste)
        try:
            text = future.result(timeout=1)
        except concurrent.futures.TimeoutError:
            self._paste_timeout_error += 1
            if self._paste_timeout_error <= _MAX_PASTE_RETRY:
                print(f'Paste process failed (TimeoutError), retrying {self._paste_timeout_error}/{_MAX_PASTE_RETRY}')
                try:
                    # noinspection PyTypeChecker
                    self._root.after(100, self._process_clip_button)
                except AttributeError:
                    return self._process_clip()
            else:
                self._paste_timeout_error = 0
                error = f'Paste process failed after {_MAX_PASTE_RETRY} attempts'
                warn(error)
            return

        self._paste_timeout_error = 0
        text = text.strip()
        if text == '':
            return self._status(self._cfg.lang('clip_empty'), True)
        self._insert_in(text, True)
        self._process()

    def _get_pipeline_results(self) -> str:
        """
        Returns the pipeline results.

        :return: Text
        """
        return str(self._text_out.get(0.0, tk.END)).strip()

    def _copy_to_clip(self) -> None:
        """
        Copy results to clip.
        """
        if not self._clip:
            return
        self._status(self._cfg.lang('status_copy_to_clip'), clear=True)
        text = self._text_out.get(0.0, tk.END)
        pyperclip.copy(text)

    def _open_settings(self) -> None:
        """
        Launch settings.
        """
        if self._settings_window:
            self._settings_window.root.lift()
            return
        self._settings_window = gui_ut.SettingsWindow((420, 330), self._cfg)
        self._settings_window.on_destroy = self._close_settings
        try:
            # self._settings_window.root.mainloop(1)
            self._settings_window.root.update()
        except AttributeError:
            pass

    def _close_settings(self) -> None:
        """
        Close settings.
        """
        self._detect_language()
        if self._ready and self._settings_window.saved:
            # self._process()
            pass
        self._settings_window = None

    def _get_selected_word(self) -> str:
        """
        Get selected word.

        :return: Return the current selected word
        """
        word = ''

        # Text in
        try:
            if self._text_in.selection_get() != '' and self._root.focus_get() == self._text_in:
                text = self._text_in.get(0.0, tk.END)
                sf = self._text_in.count('1.0', 'sel.first')
                se = self._text_in.count('1.0', 'sel.last')
                if sf is None:
                    sf = (0,)
                # print([text[sf[0]:se[0]]])
                word = ut.get_phrase_from_cursor(text, sf[0], se[0] - 1)
        except tk.TclError:
            pass

        # Text out
        try:
            if self._text_out.selection_get() != '' and self._root.focus_get() == self._text_out and word == '':
                text = self._text_out.get(0.0, tk.END)
                sf = self._text_out.count('1.0', 'sel.first')
                se = self._text_out.count('1.0', 'sel.last')
                if sf is None:
                    sf = (0,)
                word = ut.get_phrase_from_cursor(text, sf[0], se[0] - 1)
        except tk.TclError:
            pass

        # Apply pipeline
        word = pip.strict(word)
        word = ut.tokenize(word)
        return word

    def _open_dictionary(self) -> None:
        """
        Launch dictionary.
        """
        selected_word = self._get_selected_word()
        if self._dictionary_window:
            if selected_word != '':
                self._dictionary_window.insert_word(selected_word)
            self._dictionary_window.root.lift()
            self._dictionary_window.set_lang(self._detected_lang_tag)
            return
        self._dictionary_window = gui_ut.DictionaryGUI((340, 300), self._cfg, self._detected_lang_tag, self)
        self._dictionary_window.insert_word(selected_word)
        self._dictionary_window.on_destroy = self._close_dictionary
        try:
            self._dictionary_window.root.update()
        except (AttributeError, tk.TclError):
            pass

    def _close_dictionary(self) -> None:
        """
        Close dictionary.
        """
        self._dictionary_window = None

    def _close(self) -> None:
        """
        Close the window.
        """
        if self._settings_window:
            self._settings_window.close()
        if self._dictionary_window:
            self._dictionary_window.close()
        self._root.destroy()

    def _about(self) -> None:
        """
        Show about window.
        """
        _, _, ver = self._check_version()
        # f'{self._cfg.lang("about_author")}: {pydetex.__author__}\n\n' \
        n_app_opened = self._cfg.get(self._cfg.CFG_TOTAL_OPENED_APP)
        n_word_processed = self._cfg.get(self._cfg.CFG_TOTAL_PROCESSED_WORDS)
        msg = f'PyDetex v{pydetex.version.ver}\n' \
              f'{ver}\n\n' \
              f'{self._cfg.lang("about_opened")}: {ut.format_number_d(n_app_opened, self._cfg.lang("format_d"))}\n' \
              f'{self._cfg.lang("about_processed")}: {ut.format_number_d(n_word_processed, self._cfg.lang("format_d"))}\n\n' \
              f'{pydetex.__copyright__}'

        messagebox.showinfo(title=self._cfg.lang('about'), message=msg)

    def _check_version(self) -> Tuple[bool, str, str]:
        """
        Check software version.

        :return: (Needs software update, new version, version text about)
        """
        is_outdated = False
        latest_version = '0.0.0'
        # noinspection PyBroadException
        try:
            is_outdated, latest_version = check_outdated('pydetex', str(pydetex.version.ver))
            if is_outdated:
                about_ver = self._cfg.lang('about_ver_upgrade').format(latest_version)
            else:
                about_ver = self._cfg.lang('about_ver_latest')
        except ValueError:
            about_ver = self._cfg.lang('about_ver_dev')
        except requests.exceptions.ConnectionError:
            about_ver = self._cfg.lang('about_ver_err_conn')
        except Exception:
            about_ver = self._cfg.lang('about_ver_err_unkn')
        return is_outdated, latest_version, about_ver

    def _check_version_event(self) -> None:
        """
        Check the version. If outdated, raise a popup.
        """
        outdated, latest, _ = self._check_version()
        if outdated:
            messagebox.showinfo(
                title=self._cfg.lang('version_upgrade_title'),
                message=self._cfg.lang('version_upgrade').format(latest)
            )


if __name__ == '__main__':
    PyDetexGUI().start()



"""
PyDetex
https://github.com/ppizarror/PyDetex

PIPELINES
Defines the pipelines which apply parsers.
"""

__all__ = [
    'simple',
    'strict',
    'strict_eqn',
    'PipelineType'
]

import pydetex.parsers as par
from pydetex.utils import ProgressBar
from typing import Callable

PipelineType = Callable


def simple(
    s: str,
    lang: str = 'en',
    show_progress: bool = False,
    replace_pydetex_tags: bool = True,
    remove_common_tags: bool = True,
    replace_single_chars_eqn: bool = True,
    **kwargs
) -> str:
    """
    The most simple pipeline ever.

    :param s: String latex
    :param lang: Language tag of the code
    :param show_progress: Show progress bar
    :param replace_pydetex_tags: Replace pydetex tags like symbols, cites
    :param remove_common_tags: Call ``remove_common_tags`` parser
    :param replace_single_chars_eqn: Replaces all single char equations
    :return: String with no latex!
    """
    if len(s) == 0:
        return s
    steps = 17
    if not replace_pydetex_tags:
        steps -= 1
    if not replace_single_chars_eqn:
        steps -= 1
    pb = kwargs.get('progressbar', ProgressBar(steps)) if show_progress else None
    s = '\n'.join(s.splitlines())  # Removes \r\n
    s = par.process_inputs(s, pb=pb)
    s = par.remove_comments(s, pb=pb)
    s = par.process_begin_document(s, pb=pb)
    s = par.simple_replace(s, pb=pb)
    s = par.process_def(s, pb=pb, replace=kwargs.get('replace_defs', False))
    if remove_common_tags:
        s = par.remove_common_tags(s, pb=pb)
    s = par.process_cite(s, pb=pb, compress_cite=kwargs.get('compress_cite', True))
    s = par.process_citeauthor(s, lang, pb=pb)
    s = par.process_ref(s, pb=pb)
    s = par.process_labels(s, pb=pb)
    s = par.process_items(s, lang, pb=pb)
    if replace_single_chars_eqn:
        s = par.process_chars_equations(s, lang, single_only=True, pb=pb)
    s = par.unicode_chars_equations(s, pb=pb)
    s = par.remove_comments(s, pb=pb)  # comments, replace tags, strip
    if replace_pydetex_tags:
        s = par.replace_pydetex_tags(s, pb=pb, **kwargs)
    s = par.strip_punctuation(s, pb=pb)
    s = par.simple_replace(s, pb=pb)
    if s[-1] == '\\':
        s = s[0:len(s) - 1]
    return s


def strict(
    s: str,
    lang: str = 'en',
    show_progress: bool = False,
    eqn_simple: bool = True,
    **kwargs
) -> str:
    """
    Apply simple + removes all commands.

    :param s: String latex
    :param lang: Language tag of the code
    :param show_progress: Show progress bar
    :param eqn_simple: If true, replace equations with a label, else, attempt to write it as-is
    :return: String with no latex!
    """
    pb = ProgressBar(steps=24) if show_progress else None
    if 'progressbar' not in kwargs.keys():
        # noinspection PyTypeChecker
        kwargs['progressbar'] = pb
    s = simple(s, lang, replace_pydetex_tags=False, remove_common_tags=False,
               show_progress=show_progress, replace_single_chars_eqn=False, **kwargs)  # 15 steps
    s = par.process_chars_equations(s, lang, single_only=not eqn_simple, pb=pb)
    s = par.remove_equations(s, pb=pb)
    s = par.remove_environments(s, pb=pb)
    s = par.remove_commands_param(s, lang, pb=pb)
    s = par.remove_commands_param_noargv(s, pb=pb)
    s = par.remove_comments(s, pb=pb)
    s = par.replace_pydetex_tags(s, pb=pb, **kwargs)
    s = par.strip_punctuation(s, pb=pb)
    s = par.simple_replace(s, pb=pb)
    return s


def strict_eqn(
    s: str,
    lang: str = 'en',
    show_progress: bool = False,
    **kwargs
) -> str:
    """
    Same as strict, but replaces the equations with their string representation.

    :param s: String latex
    :param lang: Language tag of the code
    :param show_progress: Show progress bar
    :return: String with no latex!
    """
    return strict(s, lang, show_progress, eqn_simple=False, **kwargs)



"""
PyDetex
https://github.com/ppizarror/PyDetex

UTILS TEX
Latex utils.
"""

__all__ = [
    'apply_tag_between_inside_char_command',
    'apply_tag_tex_commands',
    'apply_tag_tex_commands_no_argv',
    'find_tex_command_char',
    'find_tex_commands',
    'find_tex_commands_noargv',
    'find_tex_environments',
    'get_tex_commands_args',
    'TEX_COMMAND_CHARS',
    'TEX_EQUATION_CHARS',
    'tex_to_unicode'
]

import flatlatex
import os
import re

from flatlatex.parser import LatexSyntaxError
from typing import Tuple, Union, List, Dict, Optional, Any

# Flat latex object
_FLATLATEX = flatlatex.converter(ignore_newlines=False, keep_spaces=True)

# Tex to unicode
_TEX_TO_UNICODE: Dict[str, Union[Dict[Any, str], List[Tuple[str, str]]]] = {
    'latex_symbols': [],
    'subscripts': {},
    'superscripts': {},
    'textbb': {},
    'textbf': {},
    'textcal': {},
    'textfrak': {},
    'textit': {},
    'textmono': {}
}

# Valid command chars
TEX_COMMAND_CHARS = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l',
                     'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
                     'y', 'z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
                     'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V',
                     'W', 'X', 'Y', 'Z', '*', '@']
TEX_EQUATION_CHARS = [
    ('$', '$', True),
    ('\(', '\)', False),
    ('\[', '\]', False),
    ('\\begin{align*}', '\end{align*}', False),
    ('\\begin{align}', '\end{align}', False),
    ('\\begin{displaymath}', '\end{displaymath}', False),
    ('\\begin{equation*}', '\\end{equation*}', False),
    ('\\begin{equation}', '\\end{equation}', False),
    ('\\begin{gather*}', '\\end{gather*}', False),
    ('\\begin{gather}', '\\end{gather}', False),
    ('\\begin{math}', '\end{math}', False)
]


def find_tex_command_char(
    s: str,
    symbols_char: List[Tuple[str, str, bool]],
) -> Tuple[Tuple[int, int, int, int], ...]:
    """
    Find symbols command positions.

    Example:

    .. code-block:: none

               00000000001111111111....
               01234567890123456789....
        Input: This is a $formula$ and this is not.
        Output: ((10, 11, 17, 18), ...)

    :param s: Latex string code
    :param symbols_char: Symbols to check ``[(initial, final, ignore escape), ...]``
    :return: Positions
    """
    assert isinstance(symbols_char, list)
    max_len = 0
    for j in symbols_char:
        assert len(j) == 3, f'Format is (initial, final, ignore escape); but received {j}'
        assert isinstance(j[0], str) and len(j[0]) > 0 and ' ' not in j[0]
        assert isinstance(j[1], str) and len(j[1]) > 0 and ' ' not in j[1]
        assert isinstance(j[2], bool)
        max_len = max(max_len, len(j[0]), len(j[1]))

    def _find(k: int, y: int, p: bool = True) -> bool:
        """
        Returns true if from k char (in s) the symbols-char-y element is present.

        :param k: Position to start
        :param y: Indes of the symbol within the list
        :param p: Reads the first (True) or last element
        :return: True if exist
        """
        if y < 0:
            return False
        n, m, ignore_escape = symbols_char[y]
        nm = n if p else m
        total = 0
        for z in range(len(nm)):
            if s[k + z] == nm[z] and (z == 0 and (not ignore_escape or ignore_escape and s[k - 1] != '\\') or z > 0):
                total += 1
        return total == len(nm)

    def _find_initial(k: int) -> int:
        """
        Find which symbol is contained.

        :param k: Position to start from
        :return: The index of the symbol within the list
        """
        for y in range(len(symbols_char)):
            if _find(k, y):
                return y
        return -1

    s = '_' + s + ' ' * max_len
    r = False  # Inside tag
    r_u = -1
    a = 0
    found = []

    for i in range(1, len(s) - max_len):
        u = _find_initial(i)
        v = _find(i, r_u, False)
        # Open tag
        if not r and u >= 0:
            a = i
            r = True
            r_u = u
        # Close
        elif r and v:
            r = False
            f, g = a - 1, i - 1
            found.append((f, f + len(symbols_char[r_u][0]), g - 1, g + len(symbols_char[r_u][1]) - 1))

    return tuple(found)


def apply_tag_between_inside_char_command(
    s: str,
    symbols_char: List[Tuple[str, str, bool]],
    tags: Union[Tuple[str, str, str, str], str]
) -> str:
    """
    Apply tag between symbols.

    For example, if symbols are ``($, $)`` and tag is ``[1,2,3,4]``:

    .. code-block:: none

        Input: This is a $formula$ and this is not.
        Output: This is a 1$2formula3$4 and this is not

    :param s: Latex string code
    :param symbols_char:  ``[(initial, final, ignore escape), ...]``
    :param tags: Tags to replace
    :return: String with tags
    """
    if isinstance(tags, str):
        if tags == '':
            return s
        tags = (tags, tags, tags, tags)

    assert len(tags) == 4
    a, b, c, d = tags
    tex_tags = find_tex_command_char(s, symbols_char)

    if len(tex_tags) == 0:
        return s
    new_s = ''
    k = 0  # Moves through tags
    for i in range(len(s)):
        if k < len(tex_tags):
            if i == tex_tags[k][0]:
                new_s += a + s[i]
                continue
            elif tex_tags[k][0] < i < tex_tags[k][1]:
                pass
            elif i == tex_tags[k][1] and tex_tags[k][1] != tex_tags[k][3]:
                new_s += b + s[i]
                if tex_tags[k][2] - tex_tags[k][1] == 0:
                    new_s += c
                continue
            elif i == tex_tags[k][2] and tex_tags[k][2] != tex_tags[k][0]:
                new_s += s[i] + c
                continue
            elif tex_tags[k][2] < i < tex_tags[k][3]:
                pass
            elif i == tex_tags[k][3]:
                new_s += s[i] + d
                k += 1
                continue
        new_s += s[i]

    return new_s


def find_tex_commands(s: str, offset: int = 0) -> Tuple[Tuple[int, int, int, int, bool], ...]:
    """
    Find all tex commands within a code.

    .. code-block:: none

                 00000000001111111111222
                 01234567890123456789012
                         a        b c  d
        Example: This is \\aCommand{nice}...
        Output: ((8, 16, 18, 21), ...)

    :param s: Latex string code
    :param offset: Offset added to the positioning, useful when using recursive calling on substrings
    :return: Tuple if found codes ``(a, b, c, d, command continues)``
    """
    found: List = []
    is_cmd = False
    is_argv = False
    s += '_'
    a, b, c0, c1, d = 0, -1, 0, 0, 0
    depth_0 = 0  # {}
    depth_1 = 0  # []
    cont_chars = ('{', '[', ' ', '\n')
    cmd_idx = 0  # index
    mode_arg = -1

    for i in range(len(s) - 1):
        # Start a command
        if not is_cmd and s[i] == '\\' and s[i + 1] in TEX_COMMAND_CHARS:
            a, b, is_cmd, is_argv = i, -1, True, False
            cmd_idx += 1
            mode_arg = -1
            depth_0, depth_1 = 0, 0

        # If command before args encounter an invalid chad, disables the command
        elif is_cmd and not is_argv and s[i] not in cont_chars and s[i] not in TEX_COMMAND_CHARS:
            is_cmd = False
            if s[i] == '\\' and s[i + 1] in TEX_COMMAND_CHARS:
                a, b, is_cmd, is_argv = i, -1, True, False
                cmd_idx += 1

        # If command has a new line, but following chars are not space
        elif is_cmd and not is_argv and s[i] == '\n' and s[i + 1] in TEX_COMMAND_CHARS:
            is_cmd = False

        # If command, not arg, but an invalid char follows the space, disables the command
        elif is_cmd and not is_argv and s[i - 1] == ' ' and s[i] not in cont_chars:
            is_cmd = False

        # Inits a new arg
        elif is_cmd and s[i] in ('{', '[') and s[i - 1] != '\\':
            is_argv = True
            if b == -1:
                b = i - 1
                depth_0, depth_1 = 0, 0
            if s[i] == '{':
                if depth_0 == 0:
                    c0 = i + 1
                    if mode_arg < 0:
                        mode_arg = 0
                depth_0 += 1
            else:
                if depth_1 == 0:
                    c1 = i + 1
                    if mode_arg < 0:
                        mode_arg = 1
                depth_1 += 1

        # Ends the argument, only if depth condition satisfies
        elif is_cmd and is_argv and s[i] in ('}', ']') and s[i - 1] != '\\':
            if s[i] == '}':
                depth_0 -= 1
            else:  # ]
                depth_1 -= 1

            if (depth_0 == 0 and mode_arg == 0) or (depth_1 == 0 and mode_arg == 1):  # Finished
                d = i - 1
                found.append([a, b, c0 if s[i] == '}' else c1, d, cmd_idx])
                if s[i + 1] not in cont_chars:
                    is_cmd = False
                is_argv = False
                mode_arg = -1
            # elif depth_0 < 0 or depth_1 < 0:  # Invalid argument (parenthesis imbalance)
            #     is_cmd = False
            #     is_argv = False
            # mode_arg = -1

    # Add the offsets
    for f in found:
        f[0] += offset
        f[1] += offset
        f[2] += offset
        f[3] += offset

    # Check if command continues
    if len(found) == 0:
        return ()
    elif len(found) == 1:
        found[0][4] = False
    else:
        for k in range(1, len(found)):
            if found[k][4] == found[k - 1][4]:
                found[k - 1][4] = True
            else:
                found[k - 1][4] = False
            if k == len(found) - 1:
                found[k][4] = False
    for k in range(len(found)):
        # noinspection PyUnresolvedReferences
        found[k] = tuple(found[k])

    return tuple(found)


def _find_tex_env_recursive(original_s: str, s: str, offset: int = 0, depth: int = 0) -> List:
    """
    Find all environments.

    :param s: Latex string code
    :param offset: Offset applied to the search
    :return: Tuple of all commands
    """
    tags = find_tex_commands(s, offset=offset)
    new_tags = []
    for t in tags:
        a, b, c, d, _ = t
        source_cmd = s[a - offset:b - offset + 1]
        if 'begin' not in source_cmd and 'end' not in source_cmd:
            # Get the arguments of the command, and check more environments there
            cmd_args = s[c - offset:d - offset + 1]
            if 'begin' in cmd_args or 'end' in cmd_args:
                if 'newenvironment' in source_cmd or 'newcommand' in source_cmd:  # Prone to bugs
                    continue
                for tr in _find_tex_env_recursive(original_s, cmd_args, offset=c, depth=depth + 1):
                    new_tags.append(tr)
        else:
            new_tags.append(t)
    return new_tags


def find_tex_environments(s: str) -> Tuple[Tuple[str, int, int, int, int, str, int, int], ...]:
    """
    Find all tex commands within a code.

    Example:

    .. code-block:: none

                 0000000000111111111122222222223333333333
                 0123456789012345678901234567890123456789
                         a           b        c         d
        Example: This is \begin{nice}[cmd]my...\end{nice}
        Output: (('nice', 8, 20, 29, 39, 'parentenv', 0, -1), ...)

    This method also returns the name of the parent environment, the depth of the
    environment, and the depth of the item enviroment (if itemizable).

    :param s: Latex string code
    :return: Tuple if found environment ``(env_name, a, b, c, d, parent_env_name, env_depth, env_item_depth)``
    """

    def _env_common(e: str) -> str:
        """
        Return the common environment for a given name.

        :param e: Environment name
        :return: Common environment
        """
        if ('itemize' in e) or ('enumerate' in e) or ('tablenotes' in e):
            return 'item_'
        return ''

    tags = _find_tex_env_recursive(s, s)
    envs = []
    env: Dict[str, List[Tuple[int, int, str, int]]] = {}
    last_env = ''
    env_depth = 0
    cmds_cont = []
    env_depths: Dict[str, int] = {}

    for t in tags:
        a, b, c, d, _ = t
        if 'begin' in s[a:b + 1]:
            env_name = s[c:d + 1]
            c_env_name = _env_common(env_name)  # Common environment name
            if c_env_name not in env_depths.keys():
                env_depths[c_env_name] = 0
            else:
                env_depths[c_env_name] += 1
            env_i = (a, d + 2, last_env, env_depth)
            if env_name not in env:
                env[env_name] = [env_i]
            else:
                env[env_name].append(env_i)
            if a not in cmds_cont:
                cmds_cont.append(a)
                last_env = env_name
                env_depth += 1
        elif 'end' in s[a:b + 1]:
            env_name = s[c:d + 1]
            c_env_name = _env_common(env_name)  # Common environment name

            if env_name in env.keys():
                env_i = env[env_name].pop()

                # Update env itemize depth
                env_depth_item = -1
                if c_env_name != '':
                    env_depth_item = env_depths[c_env_name]
                    env_depths[c_env_name] -= 1

                envs.append((
                    env_name,  # Environment name
                    env_i[0],  # a-position of the env
                    env_i[1],  # b-position
                    a,  # c-position
                    d,  # d-position
                    env_i[2],  # parent environment name
                    env_i[3],  # depth of the environment
                    env_depth_item  # itemize depth
                ))

                if len(env[env_name]) == 0:
                    del env[env_name]
                last_env = env_i[2]
                env_depth -= 1

    return tuple(envs)


def get_tex_commands_args(
    s: str,
    pos: bool = False
) -> Tuple[Tuple[Union[str, Tuple[str, bool], Tuple[int, int]], ...], ...]:
    """
    Get all the arguments from a tex command. Each command argument has a boolean
    indicating if that is optional or not.

    .. code-block:: none

        Example: This is \aCommand[\label{}]{nice} and...
        Output: (('aCommand', ('\label{}', True), ('nice', False)), ...)

    :param s: Latex string code
    :param pos: Add the numerical position of the original string at the last position
    :return: Arguments
    """
    tags = find_tex_commands(s)
    commands = []
    command = []
    for t in tags:
        a, b, c, d, cont = t
        if len(command) == 0:
            command.append(s[a + 1:b + 1].strip())
        arg = s[c - 1:d + 2]
        optional = '[' in arg
        command.append((arg[1:-1], optional))
        if not cont:
            if pos:
                command.append((a, d + 2))
            commands.append(tuple(command))
            command = []
    return tuple(commands)


def find_tex_commands_noargv(s: str) -> Tuple[Tuple[int, int], ...]:
    """
    Find all tex commands with no arguments within a code.

    .. code-block:: none

                 00000000001111111111222
                 01234567890123456789012
                         x       x
        Example: This is \aCommand ...
        Output: ((8,16), ...)

    :param s: Latex string code
    :return: Tuple if found codes
    """
    found = []
    is_cmd = False
    s += '_'
    a = 0
    cont_chars = ('{', '[', ' ')

    for i in range(len(s) - 1):
        if not is_cmd and s[i] == '\\' and s[i + 1] in TEX_COMMAND_CHARS:
            if i > 0 and s[i - 1] == '⇲':
                continue
            a = i
            is_cmd = True

        elif is_cmd and s[i] == '\\':
            if i - 1 - a > 0:
                found.append([a, i - 1])
            a = i

        elif is_cmd and s[i] in ('{', '['):
            is_cmd = False

        # If command, not arg, but an invalid char follows the space, disables the command
        elif is_cmd and s[i - 1] == ' ' and s[i] not in cont_chars:
            is_cmd = False
            found.append([a, i - 1])

        elif is_cmd and s[i] not in TEX_COMMAND_CHARS and s[i] not in cont_chars:
            is_cmd = False
            found.append([a, i - 1])

    if is_cmd and a != len(s) - 2:
        found.append([a, len(s) - 2])

    # Strip chars
    for k in range(len(found)):
        ch = found[k][1]
        for j in range(ch):
            if s[found[k][1]] == ' ':
                found[k][1] -= 1
            else:
                break
        # noinspection PyUnresolvedReferences
        found[k] = tuple(found[k])

    return tuple(found)


def apply_tag_tex_commands(
    s: str,
    tags: Union[Tuple[str, str, str, str, str], str]
) -> str:
    """
    Apply tag to tex command.

    For example, if tag is ``[1,2,3,4,5]``:

    .. code-block:: none

        Input: This is a \\formula{epic} and this is not
        Output: This is a 1\\formula2{3epic4}5 and this is not

    :param s: Latex string code
    :param tags: Tags (length 5)
    :return: Code with tags
    """
    if isinstance(tags, str):
        if tags == '':
            return s
        tags = (tags, tags, tags, tags, tags)
    assert len(tags) == 5
    a, b, c, d, e = tags  # Unpack

    tex_tags = find_tex_commands(s)
    if len(tex_tags) == 0:
        return s
    new_s = ''
    k = 0  # Moves through tags
    i = -1
    for _ in range(len(s)):
        i += 1
        if i == len(s):
            break
        if k < len(tex_tags) and i in tex_tags[k][0:4]:
            if i == tex_tags[k][0]:
                new_s += a + s[i]
            elif i == tex_tags[k][1]:
                new_s += s[i] + b
            elif i == tex_tags[k][2] and i != tex_tags[k][3]:
                new_s += c + s[i]
            elif i == tex_tags[k][3]:
                if i == tex_tags[k][2]:
                    new_s += c
                new_s += s[i] + d + s[i + 1] + e
                i += 1
                # if continues
                if tex_tags[k][4]:
                    new_s += b
                k += 1
        else:
            new_s += s[i]

    return new_s[0:len(new_s)]


def apply_tag_tex_commands_no_argv(
    s: str,
    tags: Union[Tuple[str, str], str]
) -> str:
    """
    Apply tag to tex command.

    For example, if tag is ``[1,2]``:

    .. code-block:: none

        Input: This is a \\formula and this is not.
        Output: This is a 1\\formula2 and this is not

    :param s: Latex string code
    :param tags: Tags (length 5)
    :return: Code with tags
    """
    if isinstance(tags, str):
        if tags == '':
            return s
        tags = (tags, tags)
    assert len(tags) == 2
    a, b = tags  # Unpack

    tex_tags = find_tex_commands_noargv(s)
    if len(tex_tags) == 0:
        return s
    new_s = ''
    k = 0  # Moves through tags
    i = -1
    for _ in range(len(s)):
        i += 1
        if k < len(tex_tags) and i in tex_tags[k]:
            if i == tex_tags[k][0]:
                new_s += a + s[i]
            elif i == tex_tags[k][1]:
                new_s += s[i] + b
                k += 1
        else:
            new_s += s[i]

    return new_s


def _convert_single_symbol(s: str) -> Optional[str]:
    """
    If ``s`` is just a latex code ``'alpha'`` or ``'beta'`` it converts it to its
    unicode representation.

    :param s: Latex string code
    :return: Latex with converted single symbols
    """
    if '\\' not in s[0]:
        s = '\\' + s
    for (code, val) in _TEX_TO_UNICODE['latex_symbols']:
        if code == s:
            return val
    return None


def _convert_latex_symbols(s: str) -> str:
    """
    Replace each ``'\alpha'``, ``'\beta'`` and similar latex symbols with
    their unicode representation.

    :param s: Latex string code
    :return: Replaced symbols
    """
    for (code, val) in _TEX_TO_UNICODE['latex_symbols']:
        s = s.replace(code, val)
    return s


def _process_starting_modifiers(s: str) -> str:
    """
    If s start with ``'it '``, ``'cal '``, etc. then make the whole string
    italic, calligraphic, etc.

    :param s: Latex string code
    :return: Modified text
    """
    s = re.sub('^bb ', r'\\bb{', s)
    s = re.sub('^bf ', r'\\bf{', s)
    s = re.sub('^it ', r'\\it{', s)
    s = re.sub('^cal ', r'\\cal{', s)
    s = re.sub('^frak ', r'\\frak{', s)
    s = re.sub('^mono ', r'\\mono{', s)
    return s


def _apply_all_modifiers(s: str) -> str:
    """
    Applies all modifiers.

    :param s: Latex string code
    :return: Text with replaced chars
    """
    s = _apply_modifier(s, '^', _TEX_TO_UNICODE['superscripts'])
    s = _apply_modifier(s, '_', _TEX_TO_UNICODE['subscripts'])

    s = _apply_modifier(s, '\\bb', _TEX_TO_UNICODE['textbb'])
    s = _apply_modifier(s, '\\bf', _TEX_TO_UNICODE['textbf'])
    s = _apply_modifier(s, '\\cal', _TEX_TO_UNICODE['textcal'])
    s = _apply_modifier(s, '\\emph', _TEX_TO_UNICODE['textit'])
    s = _apply_modifier(s, '\\frak', _TEX_TO_UNICODE['textfrak'])
    s = _apply_modifier(s, '\\it', _TEX_TO_UNICODE['textit'])
    s = _apply_modifier(s, '\\mono', _TEX_TO_UNICODE['textmono'])

    return s


def _apply_modifier(s: str, modifier: str, d: Dict[Any, str]) -> str:
    """
    This will search for the ^ signs and replace the next
    digit or (digits when {} is used) with its/their uppercase representation.

    :param s: Latex string code
    :param modifier: Modifier command
    :param d: Dict to look upon
    :return: New text with replaced text.
    """
    s = s.replace(modifier, "^")
    newtext = ""
    mode_normal, mode_modified, mode_long = range(3)
    mode = mode_normal
    for ch in s:
        if mode == mode_normal and ch == '^':
            mode = mode_modified
            continue
        elif mode == mode_modified and ch == '{':
            mode = mode_long
            continue
        elif mode == mode_modified:
            newtext += d.get(ch, ch)
            mode = mode_normal
            continue
        elif mode == mode_long and ch == '}':
            mode = mode_normal
            continue

        if mode == mode_normal:
            newtext += ch
        else:
            newtext += d.get(ch, ch)
    return newtext


def __load_unicode() -> None:
    """
    Loads the unicode data.
    """
    respath = str(os.path.abspath(os.path.dirname(__file__))).replace('\\', '/') + '/res/u_'
    for j in _TEX_TO_UNICODE.keys():
        if j == 'latex_symbols':
            with open(f'{respath}symbols.txt', 'r', encoding='utf-8') as f:
                line = f.readline()
                while line != "":
                    words = line.split()
                    code = words[0]
                    val = words[1]
                    _TEX_TO_UNICODE['latex_symbols'].append((code, val))
                    line = f.readline()
        else:
            with open(f'{respath}{j}.txt', 'r', encoding='utf-8') as f:
                line = f.readline()
                while line != '':
                    words = line.split()
                    code = words[0]
                    val = words[1]
                    _TEX_TO_UNICODE[j][code] = val
                    line = f.readline()


def tex_to_unicode(s: str) -> str:
    """
    Transforms tex code to unicode.

    :param s: Latex string code
    :return: Text in unicode
    """
    if s.strip() == '':
        return s
    ss = _convert_single_symbol(s)
    if ss is not None:
        return ss

    s = _convert_latex_symbols(s)
    s = _process_starting_modifiers(s)
    s = _apply_all_modifiers(s)

    # Last filter
    s = s.replace('\n\n', '\n').replace('  ', ' ').replace('\t', ' ')
    try:
        s = _FLATLATEX.convert(s)
    except LatexSyntaxError:
        pass

    return s


# Loads the unicode data
__load_unicode()



"""
PyDetex
https://github.com/ppizarror/PyDetex

UTILS
Module that contain all util methods and classes used in parsers and pipelines,
from tex, language, and low-level.
"""

__all__ = [
    'apply_tag_between_inside_char_command',
    'apply_tag_tex_commands',
    'apply_tag_tex_commands_no_argv',
    'Button',
    'check_repeated_words',
    'complete_langs_dict',
    'detect_language',
    'find_tex_command_char',
    'find_tex_commands',
    'find_tex_commands_noargv',
    'find_tex_environments',
    'format_number_d',
    'get_diff_startend_word',
    'get_language_name',
    'get_local_path',
    'get_number_of_day',
    'get_tex_commands_args',
    'get_word_from_cursor',
    'IS_OSX',
    'LangTexTextTags',
    'make_stemmer',
    'open_file',
    'ProgressBar',
    'RESOURCES_PATH',
    'split_tags',
    'syntax_highlight',
    'TEX_COMMAND_CHARS',
    'TEX_EQUATION_CHARS',
    'tex_to_unicode',
    'tokenize',
    'validate_float',
    'validate_int'
]

import datetime
import os
import platform
import sys
import time

from pathlib import Path
from typing import List, Tuple, Dict

from pydetex._fonts import FONT_TAGS as _FONT_TAGS
from pydetex._utils_lang import *
from pydetex._utils_tex import *

# Resources path
__actualpath = str(os.path.abspath(os.path.dirname(__file__))).replace('\\', '/') + '/'
RESOURCES_PATH = __actualpath + 'res/'

# Check OS
IS_OSX = platform.system() == 'Darwin'

# Import Button widget
if IS_OSX:
    from tkmacosx import Button
else:
    from tkinter import Button


def split_tags(s: str, tags: List[str]) -> List[Tuple[str, str]]:
    """
    Split a string based on tags, each line is then tagged.

    String format:
    [TAG1]new line[TAG2]this is[TAG1]very epic

    Output:
    [('TAG1', 'newline'), ('TAG', 'this is), ('TAG1', 'very epic') ... ]

    :param s: String
    :param tags: Tag list
    :return: Split tags
    """
    assert len(tags) > 0
    tagged_lines: List[Tuple[str, str]] = []
    r = 0
    for tag in tags:
        if r == 0:  # First occurence
            new = s.split(tag)
            for j in new:
                if j == '':
                    continue
                tagged_lines.append((tag, j))
        else:
            new_tagged_lines: List[Tuple[str, str]] = []
            for j in range(len(tagged_lines)):
                if tag in tagged_lines[j][1]:  # If tag exists
                    new = tagged_lines[j][1].split(tag)
                    new_tagged_lines.append((tagged_lines[j][0], new[0]))
                    for w in range(len(new) - 1):
                        new_tagged_lines.append((tag, new[w + 1]))
                else:
                    new_tagged_lines.append(tagged_lines[j])
            tagged_lines = new_tagged_lines

        r += 1

    # Merge consecutive tags
    merged_tags: List[Tuple[str, str]] = []
    r = 0
    for tagged in tagged_lines:
        if len(merged_tags) == 0 or tagged[0] != merged_tags[r - 1][0]:
            merged_tags.append(tagged)
            r += 1
        else:
            merged_tags[r - 1] = (tagged[0], merged_tags[r - 1][1] + tagged[1])

    return merged_tags


def button_text(s: str) -> str:
    """
    Generates the button text.

    :param s: Button's text
    :return: Text
    """
    return s if IS_OSX else f'  {s}  '


def validate_int(p: str) -> bool:
    """
    Validate an integer.

    :param p: Value
    :return: True if integer
    """
    if p == '' or p == '-':
        return True
    try:
        p = float(p)
        return int(p) == p
    except ValueError:
        pass
    return False


def validate_float(p: str) -> bool:
    """
    Validate a float.

    :param p: Value
    :return: True if integer
    """
    if p == '' or p == '-':
        return True
    try:
        float(p)
        return True
    except ValueError:
        pass
    return False


def syntax_highlight(s: str) -> str:
    """
    Syntax highlighter.

    :param s: Latex string code
    :return: Code with format
    """
    # Add initial normal
    s = _FONT_TAGS['normal'] + s.strip()

    # Format equations
    s = apply_tag_between_inside_char_command(
        s=s,
        symbols_char=TEX_EQUATION_CHARS,
        tags=(_FONT_TAGS['equation_char'], _FONT_TAGS['equation_inside'],
              _FONT_TAGS['equation_char'], _FONT_TAGS['normal'])
    )

    # Format commands with {arguments}
    s = apply_tag_tex_commands(
        s=s,
        tags=(_FONT_TAGS['tex_command'],
              _FONT_TAGS['normal'],
              _FONT_TAGS['tex_argument'],
              _FONT_TAGS['normal'],
              '')
    )

    # Format commands without arguments
    s = apply_tag_tex_commands_no_argv(
        s=s,
        tags=(_FONT_TAGS['tex_command'], _FONT_TAGS['normal'])
    )

    # Return formatted string
    return s


def format_number_d(n: int, c: str) -> str:
    """
    Formats a number on thousands.

    :param n: Number
    :param c: Format char
    :return: Formatted number
    """
    assert isinstance(n, int)
    return format(n, ',').replace(',', c)


def get_number_of_day() -> int:
    """
    Return the number of the day from the current year.

    :return: Day number
    """
    return datetime.datetime.now().timetuple().tm_yday


def open_file(f: str) -> str:
    """
    Open file and return its string.

    :param f: Filename
    :return: File content
    """
    o = open(f, 'r', encoding='utf-8')
    text = ''.join(o.readlines())
    o.close()
    return text


def make_path_if_not_exists(path: str) -> str:
    """
    Create path if not exists.

    :param path: Path
    :return: Path
    """
    if not os.path.isdir(path):
        Path(path).mkdir(parents=True, exist_ok=True)
    return path


def get_local_path() -> str:
    """
    :return: Returns the app local path
    """
    appdata = os.getenv('LOCALAPPDATA')
    if appdata is None:
        appdata = os.path.join(get_user_path(), 'Applications')

    path = os.path.join(appdata, 'PyDetex')
    return make_path_if_not_exists(path)


def get_user_path() -> str:
    """
    :return: Returns the user path
    """
    return os.path.expanduser('~')


class ProgressBar(object):
    """
    Basic progress bar implementation.
    """

    _current: int
    _last_step: float
    _size: int
    _step_times: Dict[str, float]
    _steps: int
    _t0: float

    def __init__(self, steps: int, size: int = 15) -> None:
        """
        Constructor.

        :param steps: How many steps have the procedure
        :param size: Bar size
        """
        assert isinstance(steps, int) and steps >= 1
        assert isinstance(size, int) and size >= 1
        self._current = 0
        self._last_step = time.time()
        self._size = size  # Bar size
        self._step_times = {}
        self._steps = steps - 1
        self._t0 = time.time()

    def _print_progress_bar(self, i: int, max_: int, post_text: str) -> None:
        """
        Prints a progress bar.

        :param i: Progress bar
        :param max_: Max steps
        :param post_text: Status
        """
        j = i / max_
        sys.stdout.write('\r')
        sys.stdout.write(f"[{'=' * int(self._size * j):{self._size}s}] {int(100 * j)}%  {post_text}")
        sys.stdout.flush()

    def update(self, status: str = '', print_total_time: bool = True) -> None:
        """
        Update the current status to a new step.

        :param status: Status text
        :param print_total_time: Prints total computing time
        """
        if self._current > self._steps:
            return
        self._print_progress_bar(self._current, self._steps, status)
        dt = time.time() - self._last_step
        self._last_step = time.time()
        self._step_times[status] = dt
        self._current += 1
        if self._current == self._steps + 1:
            print('')
            sys.stdout.flush()
            if print_total_time:
                print(f'Process finished in {time.time() - self._t0:.3f} seconds')

    def detail_times(self) -> None:
        """
        Print times.
        """
        for k in self._step_times.keys():
            print(f'{self._step_times[k]:.3f}s\t{k}')

    def reset(self) -> None:
        """
        Reset the steps.
        """
        self._current = 0
        self._t0 = time.time()
        self._last_step = time.time()
        self._step_times.clear()
